package br.usp.ime.oneintegration;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.regex.MatchResult;

import br.usp.ime.algorithms.Algorithm;
import br.usp.ime.algorithms.FastestAlgorithm3;
import br.usp.ime.algorithms.FastestAlgorithmAlt;
import br.usp.ime.algorithms.ForemostAlgorithm;
import br.usp.ime.algorithms.ShortestAlgorithm;
import br.usp.ime.evolvinggraph.Edge;
import br.usp.ime.evolvinggraph.EvolvingGraph;
import br.usp.ime.evolvinggraph.Journey;

public class OneIntegration {

	private final static String ONE_DIR = "../one_1.3.0/";
	private static final String FOREMOST_HASH_CLASS = "ForemostJourneyHash";
	private static final String SHORTEST_HASH_CLASS = "ShortestJourneyHash";
	private static final String FASTEST_HASH_CLASS = "FastestJourneyHash";
	
	public static void main(String[] args) {
		int nodesNum = Integer.parseInt(readArg(args, 0, "126"));
		int maxTime = Integer.parseInt(readArg(args, 1, "5000"));
		String routerClass = readArg(args, 2, "EG");
		
		//String connFile = ONE_DIR + "reports/" + routerClass +
		//"_ConnectivityONEReport.txt";
		//String messFile = ONE_DIR + "reports/" + routerClass +
		//"_MessageCreationReport.txt";
		String connFile = "input/testeONE4.in";
		String messFile = "input/testeONEmess4.in";
		String foremostOutputFile = ONE_DIR + "eg/" + FOREMOST_HASH_CLASS + ".java";
		String shortestOutputFile = ONE_DIR + "eg/" + SHORTEST_HASH_CLASS + ".java";
		String fastestOutputFile = ONE_DIR + "eg/" + FASTEST_HASH_CLASS + ".java";
		
		System.out.println(nodesNum + " " + maxTime + " " + routerClass);
		System.out.println("Messages: " + messFile);
		System.out.println("Connect: " + connFile);
		
		EvolvingGraph eg = EvolvingGraph.readEvolvingGraphFromONE(connFile, nodesNum, maxTime);
		EvolvingGraph eg2 = EvolvingGraph.readEvolvingGraphFromONE(connFile, nodesNum, maxTime);
		EvolvingGraph eg3 = EvolvingGraph.readEvolvingGraphFromONE(connFile, nodesNum, maxTime);
		OneIntegration one = new OneIntegration();
		
		if (routerClass.equals("foremost") || routerClass.equals("EG")){
			System.out.println("Foremost Algorithm:");
			Algorithm foremost = new ForemostAlgorithm();
			one.parse(eg, messFile, foremostOutputFile, foremost);
		}
		
		if (routerClass.equals("shortest") || routerClass.equals("EG")){
			System.out.println("Shortest Algorithm:");
			Algorithm shortest = new ShortestAlgorithm();
			one.parse(eg2, messFile, shortestOutputFile, shortest);
		}
		
		if (routerClass.equals("fastest") || routerClass.equals("EG")){
			System.out.println("Fastest Algorithm:");
			Algorithm fastest = new FastestAlgorithmAlt();
			one.parse(eg3, messFile, fastestOutputFile, fastest);
		}
	}
	
	private static String readArg(String[] args, int pos, String standard) {
		if (args.length > pos) return args[pos];
		return standard;
	}
	
	public void parse(EvolvingGraph eg, String messageReport, String routeFile, Algorithm alg) {
		File f = new File(messageReport);
		int numMessages = 0;
		try {
			Scanner input = new Scanner(f);
			FileWriter output = new FileWriter(routeFile);
			if (alg.getClass() == ForemostAlgorithm.class)
				writeHeader(output, FOREMOST_HASH_CLASS);
			else if (alg.getClass() == ShortestAlgorithm.class)
				writeHeader(output, SHORTEST_HASH_CLASS);
			else if (alg.getClass() == FastestAlgorithm3.class || 
					alg.getClass() == FastestAlgorithmAlt.class)
				writeHeader(output, FASTEST_HASH_CLASS);
			while (input.hasNextLine()) {
				if (processLine(eg, input, output, alg))
					numMessages ++;
				input.nextLine();
			}
			writeTrailer(output);
			output.close();
		} catch (Exception e) {
			System.out.println("Parser Error:");
			e.printStackTrace();
		}
		System.out.println("Routes found for " + numMessages + " messages.");
	}
	
	public boolean processLine(EvolvingGraph eg, Scanner input, FileWriter output, 
			Algorithm alg) throws Exception {
		input.findInLine("(\\d+).0000 (\\w+) (\\d+) (\\d+)");
		MatchResult result = input.match();
		
		int time = Integer.parseInt(result.group(1));
		String name = result.group(2);
		int orig = Integer.parseInt(result.group(3));
		int dest = Integer.parseInt(result.group(4));
		
		Journey j = findJourney(eg, alg, time, orig, dest);
		if (j == null || j.getEdges().isEmpty())
			return false;
		outputJourneyToFile(output, name, j);
		eg.removeJourney(j);
		return true;
	}

	private Journey findJourney(EvolvingGraph eg, Algorithm alg, int time, int orig, int dest) {
		List<Journey> journeys = alg.calculate(eg, eg.getNode(orig), time);
		Journey j = journeys.get(dest);
		if (j != null){
			if (orig == 94 && dest == 96){
				System.out.println("custo : " + j.getTransitTime());
			}
		}
		return j;
	}

	private void outputJourneyToFile(FileWriter output, String name, Journey j) throws IOException {
		ArrayList<Edge> e = j.getEdges();
		ArrayList<Integer> t = j.getTimes();
		for (int i = 0; i < e.size(); i++) {
			output.write("		messages.put(\"" + name + 
					"N" + e.get(i).getU().getId() + 
					"T" + t.get(i) + "\", " + e.get(i).getV().getId() + ");\n");
		}
	}

	private void writeHeader(FileWriter fw, String className) throws IOException {
		fw.write("package eg;\n" +
				"import java.util.HashMap;\n" +
				"\n" +
				"public class "+ className + " {\n" +
				"	static HashMap<String, Integer> messages = " +
				"new HashMap<String, Integer>();\n\n" +
				"	static{\n");
	}

	private void writeTrailer(FileWriter fw) throws IOException {
		fw.write("	}\n\n" +
				"	public HashMap<String, Integer> getMessages() {\n" +
				"		return messages;\n" +
				"	}\n" +
				"}");
	}


}
